import os
import cv2
import numpy as np
import tkinter as tk
import customtkinter as ctk
from PIL import Image, ImageTk

class NumpyViewer3D(ctk.CTk):
    """A 3D viewer for numpy arrays.
    """

    def __init__(self, array: np.ndarray) -> None:
        
        super().__init__(fg_color="black")
        
        self.apptitle = ("3D Numpy Viewer")
        self.author = "Sebastian Schäfer"
        self.version = "1.0"
        self.title(f"{self.apptitle} v{self.version} by {self.author}")

        ctk.set_appearance_mode("dark")
        ctk.set_default_color_theme("dark-blue")

        self.geometry("800x600")
        self.minsize(800, 600)
        
        self.image_data = array
        self.dimensions = self.image_data.shape        
        self.X = array
        self.dimensions = self.X.shape
        
        self.Y = np.rot90(self.X, k=1, axes=(1, 0))
        self.Z = np.rot90(self.Y, k=1, axes=(0, 2))
        self.Z = np.rot90(self.Z, k=-1, axes=(1, 2))

        self.index = 0
        self.xindex = 0
        self.yindex = 0
        self.zindex = 0
        
        self.axis = 0

        self.tabview = ctk.CTkTabview(self)
        self.tabview.add("Single")
        self.tabview.add("Triple")

        self.single_frame = ctk.CTkFrame(self.tabview.tab("Single"), bg_color="black")
        self.triple_frame = self.TripleFrame(self.tabview.tab("Triple"))

        self.controlframe = self.ControlFrame(self.single_frame)
        self.viewframe = self.ViewFrame(self.single_frame)

        self.single_frame.columnconfigure(0, weight=1)
        self.single_frame.rowconfigure(0, weight=1)
        self.single_frame.rowconfigure(1, weight=0)

        self.viewframe.grid(row=0, column=0, sticky="nsew", padx=5, pady=5)
        self.controlframe.grid(row=1, column=0, sticky="nsew", padx=5, pady=5)

        self.single_frame.pack(fill=tk.BOTH, expand=True)
        self.triple_frame.pack(fill=tk.BOTH, expand=True)

        self.tabview.pack(fill=tk.BOTH, expand=True)
        self.single_frame.bind("<Configure>", lambda x : self.viewframe.update_image())

        self.mainloop()
        
    class ControlFrame(ctk.CTkFrame):
        """The control frame of the application.
        """
            
        def __init__(self, parent) -> None:
            super().__init__(master = parent, bg_color="black")

            self.parent = parent.master.master.master

            self.index_slider = ctk.CTkSlider(self, from_=0, to= self.parent.dimensions[0]-1, number_of_steps = self.parent.dimensions[0], orientation=tk.HORIZONTAL, width=300, height=10, command = self.update_index_label)
            self.index_slider.set(0)
            self.slider_label = ctk.CTkLabel(self, text=f"Index: {int(self.index_slider.get())}", font=("Arial", 20), text_color="white")

            self.AxisLabel = ctk.CTkLabel(self, text="Axis:", font=("Arial", 20), text_color="white")
            self.Axis = tk.IntVar(value=0)
            self.XAxis = ctk.CTkRadioButton(self, text="X", variable=self.Axis, value=0, font=("Arial", 20), text_color="white", command = self.change_axis)
            self.YAxis = ctk.CTkRadioButton(self, text="Y", variable=self.Axis, value=1, font=("Arial", 20), text_color="white", command = self.change_axis)
            self.ZAxis = ctk.CTkRadioButton(self, text="Z", variable=self.Axis, value=2, font=("Arial", 20), text_color="white", command = self.change_axis)

            self.columnconfigure(0, weight=1)
            self.columnconfigure(2, weight=1)
            self.rowconfigure(0, weight=0)

            self.index_slider.grid(row=0, column=0, sticky="nsew", padx=5, pady=5)
            self.slider_label.grid(row=0, column=1, sticky="nsew", padx=5, pady=5)
            self.AxisLabel.grid(row=0, column=3, sticky="nsew", padx=5, pady=5)
            
            self.XAxis.grid(row=0, column=4, sticky="W", padx=5, pady=5)
            self.YAxis.grid(row=0, column=5, sticky="W", padx=5, pady=5)
            self.ZAxis.grid(row=0, column=6, sticky="W", padx=5, pady=5)

        def update_index_label(self, value):
            self.slider_label.configure(text=f"Index: {int(value)}")
            setattr(self.parent, {0: "xindex", 1: "yindex", 2: "zindex"}[self.parent.axis], int(value))
            setattr(self.parent, "index", {0: self.parent.xindex, 1: self.parent.yindex, 2: self.parent.zindex}[self.parent.axis])
            self.parent.viewframe.update_image()

        def change_axis(self):
            self.parent.axis = self.Axis.get()
            setattr(self.parent, "index", {0: self.parent.xindex, 1: self.parent.yindex, 2: self.parent.zindex}[self.parent.axis])
            self.parent.image_data = {0: self.parent.X, 1: self.parent.Y, 2: self.parent.Z}[self.parent.axis]
            self.parent.dimensions = self.parent.image_data.shape
            self.index_slider.configure(to=self.parent.dimensions[0]-1, number_of_steps=self.parent.dimensions[0])
            self.index_slider.set({0: self.parent.xindex, 1: self.parent.yindex, 2: self.parent.zindex}[self.parent.axis]+1)
            self.slider_label.configure(text=f"Index: {int(self.index_slider.get())}")            
            self.parent.viewframe.update_image()

        def scroll(self, event):
            self.index_slider.set(self.index_slider.get() + event.delta/120)
            self.update_index_label(self.index_slider.get())


    class ViewFrame(ctk.CTkFrame):
        """The view frame of the application.
        """

        def __init__(self, parent) -> None:
            super().__init__(master = parent, bg_color="black")

            self.parent = parent.master.master.master

            self.columnconfigure(0, weight=1)
            self.rowconfigure(0, weight=1)

            self.image = self.array_to_image(self.parent.image_data, self.parent.index)
            self.label = tk.Label(self, bg="black", image = self.image)
            self.label.image = self.image
            self.update_image()

            self.label.grid(row=0, column=0, sticky="nsew", padx=5, pady=5)

            self.label.bind("<Enter>", self.on_enter)
            self.label.bind("<Leave>", self.on_leave)

        def on_enter(self, event):                
            self.parent.bind("<MouseWheel>", self.parent.controlframe.scroll)
        def on_leave(self, event):        
            self.parent.unbind("<MouseWheel>")

        def array_to_image(self, array: np.ndarray, slice: int) -> ctk.CTkImage:
            """Converts a numpy array to an image.
            

            Args:
                array (np.ndarray): The array to convert.
                slice (int): The slice to take.

            Returns:
                ctk.CTkImage: The image.
            """
   
            image = array[slice,:,:] 
   
            image = Image.fromarray(image)
            size = min([self.winfo_width(), self.winfo_height()])
            image = image.resize((size, size), Image.LANCZOS)

            return ImageTk.PhotoImage(image)
            

        def update_image(self):
            """Updates the image.
            """

            self.image = self.array_to_image(self.parent.image_data, self.parent.index)
            self.label.config(image = self.image)
            self.label.image = self.image

    class TripleFrame(ctk.CTkFrame):
        """Shows the three different views of the array side by side.
        """

        def __init__(self, parent) -> None:
            super().__init__(master = parent, bg_color="black")

            self.parent = parent
            self.pack_propagate(False)

            self.xview = self.ViewFrame(self, 0)
            self.yview = self.ViewFrame(self, 1)
            self.zview = self.ViewFrame(self, 2)

            self.columnconfigure(0, weight=1)
            self.columnconfigure(1, weight=1)
            self.columnconfigure(2, weight=1)
            self.rowconfigure(0, weight=1)

            self.xview.grid(row=0, column=0, sticky="nsew", padx=5, pady=5)
            self.yview.grid(row=0, column=1, sticky="nsew", padx=5, pady=5)
            self.zview.grid(row=0, column=2, sticky="nsew", padx=5, pady=5)

        class ViewFrame(ctk.CTkFrame):
            """A view frame for one of the views.
            """

            def __init__(self, parent, axis) -> None:
                super().__init__(master = parent, bg_color="black")
                self.parent = parent.master.master.master
                self.master = parent
                self.pack_propagate(False)          
                self.axis = axis
                #self.index = getattr(self.parent, {0: "xindex", 1: "yindex", 2: "zindex"}[axis])
                self.ref_dimension = getattr(self.parent, "dimensions")[axis]
                self.ref_image_data = getattr(self.parent, {0: "X", 1: "Y", 2: "Z"}[axis])
                self.index = self.ref_image_data.shape[0]//2

                self.title = ctk.CTkLabel(self, text=f"{['X', 'Y', 'Z'][axis]} view", font=("Arial", 20), text_color="white")
                self.title.pack(side="top", fill="x", expand=False, padx=5, pady=5)
                self.index_slider = ctk.CTkSlider(self, from_=0, to= self.ref_dimension-1, number_of_steps = self.ref_dimension, orientation=tk.HORIZONTAL, width=200, height=10, command = self.update_index_label)
                self.index_slider.set(self.index)
                self.slider_label = ctk.CTkLabel(self, text=f"Index: {int(self.index_slider.get())}", font=("Arial", 20), text_color="white")
                self.index_slider.pack(side="top", fill="x", expand=False, padx=5, pady=5)
                self.slider_label.pack(side="top", fill="x", expand=False, padx=5, pady=5)

                self.label = tk.Label(self, bg="black")
                self.label.pack(side="top", fill="both", expand=True, padx=5, pady=5)
                self.image = self.array_to_image(self.ref_image_data, self.index)
                self.label.config(image=self.image)
                self.label.image = self.image
                self.label.bind("<Enter>", self.on_enter)
                self.label.bind("<Leave>", self.on_leave)
                self.bind("<Configure>", lambda x: self.update_image())
                self.update_image()

            def on_enter(self, event): 
                self.parent.bind("<MouseWheel>", self.scroll)
            def on_leave(self, event):      
                self.parent.unbind("<MouseWheel>")

            def array_to_image(self, array: np.ndarray, slice: int, line_index=None, line_direction=None) -> ctk.CTkImage:
                """Converts a numpy array to an image.
                

                Args:
                    array (np.ndarray): The array to convert.
                    slice (int): The slice to take.

                Returns:
                    ctk.CTkImage: The image.
                """
    
                image = array[slice,:,:].copy()

                if line_index is not None:
                    if self.axis == 0 and line_direction =="x":
                        line_index = image.shape[1] - line_index
                    if self.axis == 1:
                        if line_direction =="x":
                            line_index = image.shape[1] - line_index
                        if line_direction =="y":
                            line_index = image.shape[0] - line_index
                    
                    if line_direction == "x":
                        cv2.line(image, (line_index, 0), (line_index, image.shape[0]), (255, 0, 0), 1)
                    elif line_direction == "y":
                        cv2.line(image, (0, line_index), (image.shape[1], line_index), (255, 0, 0), 1)

                image = Image.fromarray(image)
                size = min([self.label.winfo_height(), self.winfo_width()])
                image = image.resize((size,size), Image.LANCZOS)

                return ImageTk.PhotoImage(image)
                
            def update_image(self, line_index=None, line_direction=None):
                """Updates the image.
                """

                self.image = self.array_to_image(self.ref_image_data, self.index,  line_index=line_index, line_direction=line_direction,)
                self.label.config(image = self.image)
                self.label.image = self.image

            def update_index_label(self, value):
                    
                self.slider_label.configure(text=f"Index: {int(value)}")
                self.index = int(value)
                setattr(self.parent, {0: "xindex", 1: "yindex", 2: "zindex"}[self.axis], self.index+1)
                getattr(self.master, {0: "yview", 1: "zview", 2: "xview"}[self.axis]).update_image(self.index, {0: "y", 1: "y", 2: "x"}[self.axis])
                getattr(self.master, {0: "zview", 1: "xview", 2: "yview"}[self.axis]).update_image(self.index, {0: "x", 1: "y", 2: "x"}[self.axis])
                self.update_image()

            def scroll(self, event):
                self.index_slider.set(self.index_slider.get() + event.delta/120)
                self.update_index_label(self.index_slider.get())


import cv2
import imutils
import pydicom
import numpy as np
import pandas as pd

from dataclasses import dataclass

@dataclass
class DICOMBeamMLC:
    """Class to represent the MLC Sequence for a beam.
    """

    def __init__(self, control_points:int, collimator_angle:float, mlc_leaf_sequence: dict, gantry_angles:list) -> None:
        """Initialize the DICOMBeamMLC class.

        Args:
            control_points (int): Number of control points.
            collimator_angle (float): Collimator angle.
            mlc_leaf_sequence (dict): MLC leaf sequence.
            gantry_angles (list): Gantry angles.
        """

        self._control_points = control_points
        self._collimator_angle = collimator_angle
        self._mlc_leaf_sequence = mlc_leaf_sequence	
        self._gantry_angles = gantry_angles


    def getNumberOfControlPoints(self) -> int:
        """Get the number of control points.

        Returns:
            int: Number of control points.
        """

        return self._control_points

    def getCollimatorAngle(self) -> float:
        """Get the collimator angle.

        Returns:
            float: Collimator angle.
        """

        return self._collimator_angle

    def getMLCLeafSequence(self) -> dict:
        """Get the MLC leaf sequence.

        Returns:
            dict: MLC leaf sequence.
        """

        return self._mlc_leaf_sequence

    def getGantryAngles(self) -> list:
        """Get the gantry angles.

        Returns:
            list: Gantry angles.
        """

        return self._gantry_angles


class DICOMMLC:
    """Class to represent the MLC Sequence for a DICOM plan. The MLC sequence is
    a list of DICOMBeamMLC objects. The MLC can be drawn and animated. Using edge detection,
    the center of gravity of the MLC openings can be determined.
    """

    def __init__(self, RTPlan) -> None:
        """Initialize the DICOMMLC class.

        Args:
            RTPlan (pathlike or pydicom.FileDataset): DICOM dataset.
        """

        if type(RTPlan) == str:
            ds = pydicom.dcmread(RTPlan)
        elif type(RTPlan) == pydicom.FileDataset:
            ds = RTPlan
        else:
            raise TypeError("The input must be a path to the DICOM RTPLAN file or the dataset containing the DICOM information.")
        
        if ds.Modality != "RTPLAN":
            raise TypeError("The file is not a DICOM RTPLAN file")


        self._number_of_beams = len(ds.BeamSequence)

        for i in range(len(ds.BeamSequence[0].BeamLimitingDeviceSequence)):
            if ds.BeamSequence[0].BeamLimitingDeviceSequence[i].RTBeamLimitingDeviceType in ["MLCX","MLCY"] :
                mlc_index = i
                leaf_pairs = ds.BeamSequence[0].BeamLimitingDeviceSequence[i].NumberOfLeafJawPairs
                mlc_offsets = pd.DataFrame({"Offset [mm]":ds.BeamSequence[0].BeamLimitingDeviceSequence[i].LeafPositionBoundaries})
                break
        self._leaf_pairs = int(leaf_pairs)
        self._leaf_position_boundaries = mlc_offsets
        self._leaf_positions = self._InitializeLeafPositions()
        self._beam_mlc_sequence = self._InitializeBeamMLCSequence(ds)

        self._image_width = 1200
        self._image_height = 1200
        self._dimensions = (self._image_width, self._image_height, 3)
        self._pixel_spacing = 4

    def __str__(self):
        """String representation of the DICOMMLC object.

        Returns:
            str: String representation.
        """

        return


    def _InitializeLeafPositions(self) -> pd.DataFrame:
        """Calculate the leaf centers from the leaf position boundaries.

        Returns:
            pd.DataFrame: Leaf positions.
        """


        leaf_positions = pd.DataFrame({"Position [mm]": [
            np.mean(
                [
                    self.getLeafPositionBoundaries()["Offset [mm]"][i],
                    self.getLeafPositionBoundaries()["Offset [mm]"][i + 1],
                ]
            )
            for i in range(len(self.getLeafPositionBoundaries()["Offset [mm]"]) - 1)
        ]})

        return leaf_positions

    def _InitializeBeamMLCSequence(self, ds) -> list:
        """Initialize the beam MLC sequence.

        Args:
            ds (pydicom.FileDataset): DICOM dataset.

        Returns:
            list: Beam MLC sequence.
        """

        beam_mlc_sequence = {}

        for i in range(self._number_of_beams):
            mlc_positions = {}
            control_points = len(ds.BeamSequence[i].ControlPointSequence)
            collimator_angle = float(ds.BeamSequence[i].ControlPointSequence[0].BeamLimitingDeviceAngle)
            gantry_angles = [ds.BeamSequence[i].ControlPointSequence[j].GantryAngle for j in range(control_points)]
            for j in range(control_points):
                for mlc_index in range(len(ds.BeamSequence[i].ControlPointSequence[j].BeamLimitingDevicePositionSequence)):
                    if ds.BeamSequence[i].ControlPointSequence[j].BeamLimitingDevicePositionSequence[mlc_index].RTBeamLimitingDeviceType in ["MLCX","MLCY"] :
                        break

                LeafBank1 = []
                LeafBank2 = []
                positions = ds.BeamSequence[i].ControlPointSequence[j].BeamLimitingDevicePositionSequence[mlc_index].LeafJawPositions	

                LeafBank1 += ds.BeamSequence[i].ControlPointSequence[j].BeamLimitingDevicePositionSequence[mlc_index].LeafJawPositions[:self.getLeafPairs()]
                LeafBank2 += ds.BeamSequence[i].ControlPointSequence[j].BeamLimitingDevicePositionSequence[mlc_index].LeafJawPositions[self.getLeafPairs():]

                mlc_positions[j] = {}
                mlc_positions[j]["LeafBank1"] = pd.DataFrame({"Leaf Bank 1 [mm]":LeafBank1})	
                mlc_positions[j]["LeafBank2"] = pd.DataFrame({" Leaf Bank 2 [mm]":LeafBank2})	

            beam_mlc_sequence[i] = DICOMBeamMLC(control_points, collimator_angle, mlc_positions, gantry_angles)

        return beam_mlc_sequence

    def DrawEntireMLCSequence(self, rotate=False, draw_edges=True):
        """Draw the entire MLC sequence.

        Args:
            rotate (bool, optional): Rotate the MLC. Defaults to False.
            draw_edges (bool, optional): Draw the edges of the MLC. Defaults to True.
        """

        images = []
        for i in range(self.getNumberOfBeams()):
            for j in range(self.getBeamMLCSequence()[i].getNumberOfControlPoints()):
                images.append(self.DrawMLCAperture(i, j, rotate, draw_edges))

        return np.array(images)


    def DrawMLCAperture(self, beam, control_point, rotate=False, draw_edges=True):
        """Draw the MLC for a given beam and control point.
        
        Args:
            beam (int): Beam number.
            control_point (int): Control point number.
            rotate (bool, optional): Rotate the MLC. Defaults to False.
            draw_edges (bool, optional): Draw the edges of the MLC. Defaults to True.
        """                

        image = np.zeros(self._dimensions, np.uint8)

        center = int((image.shape[0])/2)
        collimator_angle = self.getBeamMLCSequence()[beam].getCollimatorAngle()
        mlc_length = 100 * self.getPixelSpacing()
        mlc_positions = self.getBeamMLCSequence()[beam].getMLCLeafSequence()
        
        for i, leaf in enumerate(mlc_positions[control_point]["LeafBank1"]["Leaf Bank 1 [mm]"]):

            pointA = (center + int((leaf*self.getPixelSpacing())), center + int((self.getLeafPositionBoundaries()["Offset [mm]"][i]*self.getPixelSpacing())))
            pointB = (center + int((leaf*self.getPixelSpacing())), center + int((self.getLeafPositionBoundaries()["Offset [mm]"][i+1]*self.getPixelSpacing())))
            pointC = (pointA[0]-mlc_length, pointA[1])
            pointD = (pointB[0]-mlc_length, pointB[1])
            
            cv2.rectangle(image, pointA, pointD, (255,255,255), -1)
            if draw_edges:
                cv2.line(image, pointA, pointB, (0,0,255), 1)
                cv2.line(image, pointA, pointC, (0,0,255), 1)
                cv2.line(image, pointC, pointD, (0,0,255), 1)
                cv2.line(image, pointB, pointD, (0,0,255), 1)
            
        for i, leaf in enumerate(mlc_positions[control_point]["LeafBank2"][" Leaf Bank 2 [mm]"]):

            pointA = (center + int((leaf*self.getPixelSpacing())), center + int((self.getLeafPositionBoundaries()["Offset [mm]"][i]*self.getPixelSpacing())))
            pointB = (center + int((leaf*self.getPixelSpacing())), center + int((self.getLeafPositionBoundaries()["Offset [mm]"][i+1]*self.getPixelSpacing())))
            pointC = (pointA[0]+mlc_length, pointA[1])
            pointD = (pointB[0]+mlc_length, pointB[1])
            
            cv2.rectangle(image, pointA, pointD, (255,255,255), -1)
            if draw_edges:
                cv2.line(image, pointA, pointB, (0,0,255), 1)
                cv2.line(image, pointA, pointC, (0,0,255), 1)
                cv2.line(image, pointC, pointD, (0,0,255), 1)
                cv2.line(image, pointB, pointD, (0,0,255), 1)
            
        image = np.flip(image, axis=0)
        if rotate:
            image = imutils.rotate_bound(image, -collimator_angle) 

        return np.array(image, dtype=np.uint8)
    
    def FindApertureCenters(self, beam:int, control_point:int, lower_area_bound:int, upper_area_bound:int) -> tuple:
        """Find the center of the aperture for a given beam and control point.

        Args:
            beam (int): Beam number.
            control_point (int): Control point number.
            lower_area_bound (int): Lower bound for the area of the aperture.
            upper_area_bound (int): Upper bound for the area of the aperture.

        Returns:
            tuple: Tuple containing the center of the aperture.
        """
        
        image = self.DrawMLCAperture(beam, control_point, rotate=True, draw_edges=False)
        image = cv2.cvtColor(image, cv2.COLOR_RGB2GRAY)

        #create a blank copy of the image
        blank = np.zeros(image.shape, dtype=np.uint8)
        stencil  = blank.copy().astype(np.uint8)
        edges = cv2.Canny(image,100,200)

        contours, _ = cv2.findContours(edges, 1, 2)

        centers = []
        for contour in contours:
            stencil  = np.zeros(image.shape, dtype=np.uint8)
            cv2.fillConvexPoly(stencil, contour, 100)
            counts = cv2.countNonZero(stencil)

            if counts <= upper_area_bound and counts >= lower_area_bound:
                moments = cv2.moments(contour)
                try:
                    cX = int(moments["m10"] / moments["m00"])
                    cY = int(moments["m01"] / moments["m00"])
                except ZeroDivisionError:
                    continue
                if centers == []:
                    centers += [(cX, cY)]
                    cv2.circle(stencil, (cX, cY), 5, (0, 0, 255), -1)
                #check the centers list to see if a cenetr point +-2 pixels is already in the list
                if len(centers) > 0:
                    for center in centers:
                        if abs(center[0] - cX) < 2 and abs(center[1] - cY) < 2:
                            break
                    else:
                        centers += [(cX, cY)]
                        cv2.circle(stencil, (cX, cY), 5, (0, 0, 255), -1)
                        continue
                    
        return centers

    def getNumberOfBeams(self) -> int:
        """Get the number of beams.

        Returns:
            int: Number of beams.
        """

        return self._number_of_beams


    def getLeafPairs(self) -> int:
        """Get the number of leaf pairs.

        Returns:
            int: Number of leaf pairs.
        """

        return self._leaf_pairs

    def getLeafPositionBoundaries(self) -> pd.DataFrame:
        """Get the leaf position boundaries.

        Returns:
            pd.DataFrame: Leaf position boundaries.
        """

        return self._leaf_position_boundaries

    def getLeafPositions(self) -> pd.DataFrame:
        """Get the leaf positions.

        Returns:
            pd.DataFrame: Leaf positions.
        """

        return self._leaf_positions

    def getBeamMLCSequence(self) -> list:
        """Get the beam MLC sequence.

        Returns:
            list: Beam MLC sequence.
        """

        return self._beam_mlc_sequence        

    def setImageWidth(self, width: int) -> None:
        """Set the width of the output array in pixels.

        Args:
            width (int): Width in pixels.
        """
        if width > 0:
            self._image_width = width
            self._dimensions = (self.image_width, self.image_height, 3)    
        else:
            raise ValueError("^Width must be greater than 0")
        
    def getImageWidth(self) -> int:
        """Get the width of the output array in pixels.

        Returns:
            int: Width in pixels.
        """
        return self._image_width	   
        	
    def setImageHeight(self, height: int) -> None:
        """Set the height of the output array in pixels.

        Args:
            height (int): Height in pixels.
        """
        if height > 0:
            self._image_height = height	
            self._dimensions = (self._image_width, self._image_heigh, 3)	
        else:
            raise ValueError("Height must be greater that 0")
        
    def getImageHeight(self) -> int:
        """Get the height of the output array in pixels.

        Returns:
            int: Height in pixels.
        """
        return self._image_height	
    
    def setDimensions(self, width:int, height:int) -> None:
        """Set the dimensions of the output array in pixels.

        Args:
            width (int): Width in pixels.
            height (int): Height in pixels.
        """
        if width >0 and height > 0:
            self._image_width = width	
            self._image_height = height
            self._dimensions = (self._image_width, self._image_height, 3)	
        else:
            raise ValueError("Dimensions must be greater than 0")
        
    def getDimensions(self) -> tuple:
        """Get the dimensions of the output array in pixels.

        Returns:
            tuple: Dimensions in pixels.	
        """
        return (self._dimensions[0], self._dimensions[1])	
    
    def setPixelSpacing(self, x: int) -> None:
        if x > 0:
            self._pixel_spacing = x
        else:
            raise ValueError("Pixel spacing must be greater than 0")
        
    def getPixelSpacing(self) -> int:
        """Set the pixel spacing of the output array in pixels.

        Returns:
            int: Pixel spacing in pixels.		
        """
        return self._pixel_spacing


import cv2
import pydicom
import numpy as np
from dataclasses import dataclass

@dataclass
class StructureSetContour:
    """A class to store the relevant countour information of a RTStruct structure.
    """
    
    def __init__(self, name:str, color:tuple, contours:list, slices:list) -> None:
        """Initializes the StrucutrSetContour class.

        Args:
            name (str): Name of the structure.
            color (tuple): RGB color of the structure.
            contours (list): Contours of the structure.
            slices (list): Slice indices of the structure.
        """
        
        self.Name = name
        self.Color = color
        self.ContourSequence = contours
        self.Slices = slices
    
    def getName(self) -> str:
        """Get the name of the structure.

        Returns:
            str: Name of the structure.
        """
        return self.Name	
    
    def getColor(self) -> tuple:
        """Get the color of the structure.

        Returns:
            tuple: RGB color of the structure.
        """
        return self.Color
    
    def getContours(self) -> list:
        """Get the contour sequence of the structure.

        Returns:
            list: Contour sequence of the structure.
        """
        return self.ContourSequence	
    
    def getSlices(self) -> list:
        """Get the slice indices of the structure.

        Returns:
            list: Slice indices of the structure.
        """
        return self.Slices
    
class DICOMStructureSet:   
    """A class to handle structures in DICOM RTSTRUCT files.
    Structures and corresponding contours can be extrancted and manipulated without the need for the corresponding CT.
    """ 	
    
    def __init__(self, RTStruct) -> None:
        """Initializes the DICOMStructureSet class.

        Args:
            RTStruct (pathlike or pydicom.FileDataset): The path to the DICOM RTSTRUCT file or the dataset containing the DICOM information.
        """
        
        if type(RTStruct) == str:
            ds = pydicom.dcmread(RTStruct)
        elif type(RTStruct) == pydicom.FileDataset:
            ds = RTStruct
        else:
            raise TypeError("The input must be a path to the DICOM RTSTRUCT file or the dataset containing the DICOM information.")
        
        if ds.Modality != "RTSTRUCT":
            raise TypeError("The file is not a DICOM RTSTRUCT file")

        self._image_width = 1200
        self._image_height = 1200
        self._dimensions = (self._image_width, self._image_height, 3)
        self._pixel_spacing = 2

        self._AvailableStructures = self._setAvailableStructures(ds)
        self._StructureContours = self._setStructureContours(ds)
        self._Slices = self._setSlices()
        
    def __str__(self) -> str:
        """String representation of the DICOMStructureSet object.

        Returns:
            str: String representation.
        """
        name = self.__class__.__name__
        number_of_structures = f"Number of structures: {len(self.getAvailableStructureNames())}"
        available_structures = f"Available structures: {', '.join(self.getAvailableStructureNames())}"
        
        return_string = f"{name}\n\n{number_of_structures}\n{available_structures}"
        return return_string
    
    
    def _setAvailableStructures(self, ds: pydicom.FileDataset) -> dict:
        """Initializes the dictionaty which contains the available structures.

        Args:
            ds (pydicom.FileDataset): Dataset containing the DICOM information.

        Returns:
            dict: Dictionary which containts the structures available in the structure set.
        """
        available_structures = {}
        for i in range(len(ds.StructureSetROISequence)):
            ROIName = ds.StructureSetROISequence[i].ROIName
            ROINumber = ds.StructureSetROISequence[i].ROINumber
            available_structures[ROIName] = ROINumber	
            
        return available_structures
            
    def _setStructureContours(self, ds: pydicom.FileDataset) -> dict:
        """Initializes the dictionary which contains the contours of the structures.

        Args:
            ds (pydicom.FileDataset): Dataset containing the DICOM information.

        Returns:
            dict: Dictionary which contains the contours of the structures.	
        """
        structure_contours = {}
        for i in range(len(ds.StructureSetROISequence)):
            name = ds.StructureSetROISequence[i].ROIName
            color = ds.ROIContourSequence[i].ROIDisplayColor
            contours =  [ ds.ROIContourSequence[i].ContourSequence[j].ContourData 
            for j in range(len(ds.ROIContourSequence[i].ContourSequence))]
            slices = []
            for k in range(len(contours)):      
                contours[k] = np.array(contours[k]).reshape(-1,3)
                slices += [contours[k][0][2]]
                contours[k] = contours[k][:,0:2]      
        
            structure_contours[name] = StructureSetContour(name, color, contours, slices)
        
        return structure_contours
            
    def _setSlices(self) -> dict:
        """Initializes the dictionary which translates the slice number to the index of the array.

        Returns:
            dict: Dictionary which translates the slice number to the index of the array.	
        """
        
        slices=[]
        for Structure in self.getAvailableStructureNames():
            for slice_number in self._StructureContours[Structure].getSlices():
                if slice_number not in slices:
                    slices.append(slice_number)
                    
        slices.sort()
        slice_dict = {x:i for i, x in enumerate(slices)}
        
        return slice_dict

    
    def DrawAllContours(self, fill = False) -> np.ndarray:
        """Draw all the contours of all the structures in the structure set.

        Returns:
            np.ndarray: Array of the slices with the contours of all the structures.
        """
        
        images = [np.zeros(self._dimensions, dtype=np.uint8) for i in range(len(self._Slices))]
        
        for Structure in self.getAvailableStructureNames():
            for contour, slice in zip(self._StructureContours[Structure].getContours(), self._StructureContours[Structure].getSlices()):
                points = np.array(contour, dtype=np.int32) * self.getPixelSpacing()
                points[:,0] += int(images[self._Slices[slice]].shape[1]/2)
                points[:,1] += int(images[self._Slices[slice]].shape[0]/2)
                cv2.drawContours(images[self._Slices[slice]], [points], 0, self._StructureContours[Structure].getColor(), self.getPixelSpacing()) 

        return np.array(images)
    
    def DrawStructureContours(self, Structure:str, RemoveEmptySlices:bool = True) -> np.ndarray:
        """_summary_

        Args:
            Structure (str): Name if the Structure.
            RemoveEmptySlices (bool, optional): Restrict the output array size the the slices which
            contain the structure, or return all the slices. Defaults to True.

        Returns:
            np.ndarray: Array of the slices with the contours of the structure.	
        """
            
        images = [np.zeros(self._dimensions, dtype=np.uint8) for i in range(len(self._Slices))]
        
        for contour, slice in zip(self._StructureContours[Structure].getContours(), self._StructureContours[Structure].getSlices()):
            points = np.array(contour, dtype=np.int32) * self.getPixelSpacing()
            points[:,0] += int(images[self._Slices[slice]].shape[1]/2)
            points[:,1] += int(images[self._Slices[slice]].shape[0]/2)
            cv2.drawContours(images[self._Slices[slice]], [points], 0, self._StructureContours[Structure].getColor(), self.getPixelSpacing()) 

        if RemoveEmptySlices:
            images = [image for image in images if not np.all(image == 0)]

        return np.array(images)   
        
    def DrawStructureContourSlice(self, Structure: str, Slice:int) -> np.ndarray:
        """Draw the contours of a structure of a specific slice.

        Args:
            Structure (str): Name of the structure.
            Slice (int): Slice index.

        Returns:
            np.ndarray: Array of the slice with the contours of the structure.
        """
            
        image = np.zeros(self._dimensions, dtype=np.uint8)
        
        if Structure not in self.getAvailableStructureNames():
            raise ValueError("The specified structure is not available")
        
        if Slice not in self._Slices.keys():
            raise ValueError("The specified slice is not available")
        
        for contour, slice in zip(self._StructureContours[Structure].getContours(), self._StructureContours[Structure].getSlices()):
            if self._Slices[slice] == Slice:
                points = np.array(contour, dtype=np.int32) * self.getPixelSpacing()
                points[:,0] += int(image.shape[1]/2) 
                points[:,1] += int(image.shape[0]/2)
                cv2.drawContours(image, [points], 0, self._StructureContours[Structure].getColor(), self.getPixelSpacing()) 

        return image	
        
    def getSliceIndices(self, Structure: str) -> list:
        """Get the indices of the slices that contain the specified structure.

        Args:
            Structure (str): Name of the structure.

        Returns:
            list: List of slice indices that contain the specified structure. 
        """
        indices = []
        for slice_number in self._StructureContours[Structure].getSlices():
            indices.append(self._Slices[slice_number])
        return indices
    	
    
    def setImageWidth(self, width: int) -> None:
        """Set the width of the output array in pixels.

        Args:
            width (int): Width in pixels.
        """
        if width > 0:
            self._image_width = width
            self._dimensions = (self.image_width, self.image_height, 3)    
        else:
            raise ValueError("^Width must be greater than 0")
        
    def getImageWidth(self) -> int:
        """Get the width of the output array in pixels.

        Returns:
            int: Width in pixels.
        """
        return self._image_width	   
        	
    def setImageHeight(self, height: int) -> None:
        """Set the height of the output array in pixels.

        Args:
            height (int): Height in pixels.
        """
        if height > 0:
            self._image_height = height	
            self._dimensions = (self._image_width, self._image_heigh, 3)	
        else:
            raise ValueError("Height must be greater that 0")
        
    def getImageHeight(self) -> int:
        """Get the height of the output array in pixels.

        Returns:
            int: Height in pixels.
        """
        return self._image_height	
    
    def setDimensions(self, width:int, height:int) -> None:
        """Set the dimensions of the output array in pixels.

        Args:
            width (int): Width in pixels.
            height (int): Height in pixels.
        """
        if width >0 and height > 0:
            self._image_width = width	
            self._image_height = height
            self._dimensions = (self._image_width, self._image_height, 3)	
        else:
            raise ValueError("Dimensions must be greater than 0")
        
    def getNumberOfSlices(self) -> int:
        """Get the number of slices contoured in the structure set.

        Returns:
            int: Number of slices contoured in the structure set.     
        """
        return len(self._Slices.keys())
        
    def getDimensions(self) -> tuple:
        """Get the dimensions of the output array in pixels.

        Returns:
            tuple: Dimensions in pixels.	
        """
        return (self._dimensions[0], self._dimensions[1])	
    
    def setPixelSpacing(self, x: int) -> None:
        if x > 0:
            self._pixel_spacing = x
        else:
            raise ValueError("Pixel spacing must be greater than 0")
        
    def getPixelSpacing(self) -> int:
        """Set the pixel spacing of the output array in pixels.

        Returns:
            int: Pixel spacing in pixels.		
        """
        return self._pixel_spacing	

    def getAvailableStructureNames(self) -> list:
        """Get a list of the structures in the structure set.

        Returns:
            list: All Structures in the structure set.
        """
        return list(self._AvailableStructures.keys())
    


import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="rtdicomtools",
    version="1.0.0",
    author="Sebastian Schäfer",
    author_email="se.schaefer@uke.de",
    description="",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/sebasj13/RTDICOMTools",
    project_urls={"Bug Tracker": "https://github.com/sebasj13/RTDICOMTools/issues",},
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    
    install_requires=[
        "numpy",
        "opencv-python",
        "Pillow",
        "matplotlib",
        "pandas",
        "pydicom",
        "customtkinter"
    ],
    packages=[
        "rtdicomtools"
    ],
    keywords=["medical", "physics", "DICOM", "radiotherapy","3D numpy viewer", "RTPLAN", "RTSTRUCT"],
    python_requires=">=3.8",
)

